# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Ragged Tensors.

This package defines ops for manipulating ragged tensors (`tf.RaggedTensor`),
which are tensors with non-uniform shapes.  In particular, each `RaggedTensor`
has one or more *ragged dimensions*, which are dimensions whose slices may have
different lengths.  For example, the inner (column) dimension of
`rt=[[3, 1, 4, 1], [], [5, 9, 2], [6], []]` is ragged, since the column slices
(`rt[0, :]`, ..., `rt[4, :]`) have different lengths.  For a more detailed
description of ragged tensors, see the `tf.RaggedTensor` class documentation
and the [Ragged Tensor Guide](/guides/ragged_tensor).


### Additional ops that support `RaggedTensor`

* `tf.debugging.check_numerics`
* `tf.identity`
* `tf.ones_like`
* `tf.ones_like`
* `tf.zeros_like`
* `tf.zeros_like`
* `tf.clip_by_value`
* `tf.math.abs`
* `tf.math.acos`
* `tf.math.acosh`
* `tf.math.angle`
* `tf.math.asin`
* `tf.math.asinh`
* `tf.math.atan`
* `tf.math.atanh`
* `tf.dtypes.cast`
* `tf.math.ceil`
* `tf.math.conj`
* `tf.math.cos`
* `tf.math.cosh`
* `tf.math.digamma`
* `tf.math.erf`
* `tf.math.erfc`
* `tf.math.exp`
* `tf.math.expm1`
* `tf.math.floor`
* `tf.math.imag`
* `tf.math.is_finite`
* `tf.math.is_inf`
* `tf.math.is_nan`
* `tf.math.lgamma`
* `tf.math.log`
* `tf.math.log1p`
* `tf.math.log_sigmoid`
* `tf.math.logical_not`
* `tf.math.negative`
* `tf.math.real`
* `tf.math.reciprocal`
* `tf.math.rint`
* `tf.math.round`
* `tf.math.rsqrt`
* `tf.dtypes.saturate_cast`
* `tf.math.sign`
* `tf.math.sin`
* `tf.math.sinh`
* `tf.math.sqrt`
* `tf.math.square`
* `tf.math.tan`
* `tf.io.decode_compressed`
* `tf.strings.to_number`
* `tf.strings.to_hash_bucket`
* `tf.dtypes.as_string`
* `tf.io.decode_base64`
* `tf.io.encode_base64`
* `tf.strings.regex_full_match`
* `tf.strings.regex_replace`
* `tf.strings.strip`
* `tf.strings.to_hash_bucket`
* `tf.strings.to_hash_bucket_fast`
* `tf.strings.to_hash_bucket_strong`
* `tf.strings.substr`
* `tf.strings.substr`
* `tf.strings.length`
* `tf.strings.length`
* `tf.strings.unicode_script`
* `tf.math.add_n`
* `tf.strings.join`
* `tf.math.add`
* `tf.math.atan2`
* `tf.dtypes.complex`
* `tf.div_no_nan`
* `tf.math.divide`
* `tf.math.equal`
* `tf.math.floordiv`
* `tf.floormod`
* `tf.math.greater`
* `tf.math.greater_equal`
* `tf.math.less`
* `tf.math.less_equal`
* `tf.math.logical_and`
* `tf.math.logical_or`
* `tf.math.logical_xor`
* `tf.math.maximum`
* `tf.math.minimum`
* `tf.math.multiply`
* `tf.math.not_equal`
* `tf.math.pow`
* `tf.realdiv`
* `tf.math.squared_difference`
* `tf.math.subtract`
* `tf.math.truediv`
* `tf.truncatediv`
* `tf.truncatemod`
* `tf.batch_gather`
* `tf.concat`
* `tf.expand_dims`
* `tf.expand_dims`
* `tf.gather`
* `tf.gather`
* `tf.gather_nd`
* `tf.stack`
* `tf.tile`
* `tf.where`
* `tf.math.unsorted_segment_sum`
* `tf.math.unsorted_segment_prod`
* `tf.math.unsorted_segment_min`
* `tf.math.unsorted_segment_max`
* `tf.math.unsorted_segment_mean`
* `tf.math.unsorted_segment_sqrt_n`
* `tf.math.reduce_sum`
* `tf.math.reduce_prod`
* `tf.math.reduce_min`
* `tf.math.reduce_max`
* `tf.math.reduce_mean`
* `tf.math.reduce_any`
* `tf.reduce_all`
"""

from __future__ import print_function as _print_function

from tensorflow.python.ops.ragged.ragged_factory_ops import constant
from tensorflow.python.ops.ragged.ragged_factory_ops import constant_value
from tensorflow.python.ops.ragged.ragged_functional_ops import map_flat_values
from tensorflow.python.ops.ragged.ragged_math_ops import range
from tensorflow.python.ops.ragged.ragged_tensor_value import RaggedTensorValue
from tensorflow.python.ops.ragged.segment_id_ops import row_splits_to_segment_ids
from tensorflow.python.ops.ragged.segment_id_ops import segment_ids_to_row_splits

del _print_function
